import numpy as np
from IPython.display import clear_output
import concurrent.futures as cf

def fpi_trans_single_mirror_sep(m_sep, wavelength, temp, mirror_sep_diff, layer_t, fractions, angles, stack_refractive_index):
    transmittance_vec = np.zeros_like(wavelength)
    for idx, wvl in enumerate(wavelength):
        for m_diff, frac in zip(mirror_sep_diff, fractions):
            layer_t[3] = m_sep + m_diff
            T = transfer_FPI(stack_refractive_index[idx,:], layer_t, wvl, angles[idx,:])
            transmittance_vec[idx] += frac * abs(1/T[0,0])**2
    return transmittance_vec

def mirror_dist_diff(r, R):
    return -2* ( 0.5*(2*R - np.sqrt(4*R**2 - (2*r)**2)) )

def FPI_trans_matrix_ang(mirror_sep, lam, temperature, angle_in_deg):
    lam0 = 10.5e-6 #Central wavelength used for layer thicknesses
    diameter = 40e-3 #mirror diameter
    h_max = 150e-9 #Maximum height difference between mirror center and edges
    R = h_max/2 + (diameter**2)/(8*h_max) #radius of mirror curvature
    n_points = 9 #number of descrete, different mirror separations across the mirror due to curvature

    r_area = np.linspace(0,diameter/2,n_points) #radii used for area calculations
    r_eval = r_area + np.diff(r_area)[0]/2 #radii at which the mirror separation difference is calculated
    r_area = np.delete(r_area, 0)
    r_eval = np.delete(r_eval, -1)
    area = []
    sep_diff = []
    for i in range(n_points-1):
        sep_diff.append(mirror_dist_diff(r_eval[i],R))
        if i == 0:
            area.append(np.pi*r_area[i]**2)
        else:
            area.append(np.pi*(r_area[i]**2-r_area[i-1]**2))
    total_area = sum(area)
    fraction = area/total_area

    phys_layer_thick = np.array([0.5*lam0/refrac_Ge(lam0, temperature), 0.25*lam0/refrac_ThF4(lam0),  0.25*lam0/refrac_Ge(lam0, temperature), 0, 0.25*lam0/refrac_Ge(lam0, temperature), 0.25*lam0/refrac_ThF4(lam0), 0.5*lam0/refrac_Ge(lam0, temperature)])
    transmittance = np.zeros([len(mirror_sep), len(lam)])
    n_ThF4 = refrac_ThF4(lam)
    n_ZnSe = refrac_ZnSe(lam)
    n_Ge = refrac_Ge(lam, temperature)
    n_air = np.ones_like(lam)*1.0002726# air refractive index
    stack_refractive_index = [n_Ge, n_ThF4, n_Ge, n_air, n_Ge, n_ThF4, n_Ge]
    temp = np.concatenate(stack_refractive_index, axis = 0)
    stack_refractive_index = temp.reshape((len(lam), len(stack_refractive_index)), order = 'F')

    ang_lst = np.zeros([len(lam), stack_refractive_index.shape[1] + 4])
    temp = np.deg2rad(angle_in_deg)
    ang_lst[:,(0,-1)] = np.tile(np.deg2rad(angle_in_deg),(2,1)).T
    ang_lst[:,(1,-2)] = np.tile(np.arcsin(n_air*np.sin(ang_lst[:,0])/n_ZnSe),(2,1)).T
    ang_lst[:,(2,-3)] = np.tile(np.arcsin(n_ZnSe*np.sin(ang_lst[:,1])/stack_refractive_index[:,0]),(2,1)).T
    for i in range((stack_refractive_index.shape[1]-1)//2):
        ang_lst[:,(i+3, -(i+4))] = np.tile(np.arcsin(stack_refractive_index[:,i]*np.sin(ang_lst[:,i+2])/stack_refractive_index[:,i+1]),(2,1)).T #For the middle layer i+3 and (-(i+4)) will reference the same element, but this is no problem

    lam_lst, temperature_lst, diff_lst, layer_thickness_lst, fraction_lst, ang_lst_lst = [], [], [], [], [], []
    refractive_index_lst = []
    for i in range(len(mirror_sep)):
        lam_lst.append(lam)
        temperature_lst.append(temperature)
        diff_lst.append(sep_diff)
        layer_thickness_lst.append(phys_layer_thick)
        fraction_lst.append(fraction)
        ang_lst_lst.append(ang_lst)
        refractive_index_lst.append(stack_refractive_index)

    with cf.ProcessPoolExecutor() as executor:
        results = executor.map(fpi_trans_single_mirror_sep, mirror_sep, lam_lst, temperature_lst, diff_lst, layer_thickness_lst, fraction_lst, ang_lst_lst, refractive_index_lst)

    for i, result in enumerate(results):
        transmittance[i,:] = result

    return transmittance


def refrac_ThF4(lam):
    # Handbook of Optical Constants of Solids II, 1998, page 1051
    lam = lam * 1e6
    n = 1.118 + 2.46 / lam - 2.98 / (lam**2)
    return n

def refrac_ZnSe(lam):
    # Handbook of Optical Constants of Solids II, page 737
    lam = lam * 1e6
    n = np.sqrt(4 + 1.9*lam**2/(lam**2 - 0.113))
    return n

def refrac_Ge(lam, temperature):
    # Handbook of Optical Constants of Solids I, page 465
    lam = lam * 1e6
    A = -6.04e-3 * temperature + 11.05128
    B = 9.295e-3 * temperature + 4.00536
    C = -5.392e-4 * temperature + 0.599034
    D = 4.151e-4 * temperature + 0.09145
    E = 1.51408 * temperature + 3426.5
    n = np.sqrt(A + (B * lam**2) / (lam**2 - C) + (D * lam**2) / (lam**2 - E))
    return n

def transfer_FPI(n, d, lam, list_of_angles):
    T_s = interface_s(refrac_ZnSe(lam), n[0], list_of_angles[1], list_of_angles[2]) @ ac_phase(n[0], d[0], lam, list_of_angles[2])
    for i in np.arange(1, len(n), 1):
        T_s = T_s @ interface_s(n[i-1], n[i], list_of_angles[i+1], list_of_angles[i+2]) @ ac_phase(n[i], d[i], lam, list_of_angles[i+2])
    T_s = T_s @ interface_s(n[-1], refrac_ZnSe(lam), list_of_angles[-3], list_of_angles[-2])
    T_p = interface_p(refrac_ZnSe(lam), n[0], list_of_angles[1], list_of_angles[2]) @ ac_phase(n[0], d[0], lam, list_of_angles[2])
    for i in np.arange(1, len(n), 1):
        T_p = T_p @ interface_p(n[i-1], n[i], list_of_angles[i+1], list_of_angles[i+2]) @ ac_phase(n[i], d[i], lam, list_of_angles[i+2])
    T_p = T_p @ interface_p(n[-1], refrac_ZnSe(lam), list_of_angles[-3], list_of_angles[-2])
    T = 0.5*(T_s + T_p)
    return T

def interface_s(n1, n2, angle1, angle2):
    #n1 is refractive index of leftmost film, while n2 is refracive index of rightmost film
    r_right_s = (n1*np.cos(angle1) - n2*np.cos(angle2))/(n1*np.cos(angle1) + n2*np.cos(angle2))
    r_left_s = (n2*np.cos(angle2) - n1*np.cos(angle1))/(n2*np.cos(angle2) + n1*np.cos(angle1))

    t_right_s = 2*n1*np.cos(angle1)/(n1*np.cos(angle1) + n2*np.cos(angle2))
    t_left_s = 2*n2*np.cos(angle2)/(n2*np.cos(angle2) + n1*np.cos(angle1))

    D = np.ones([2,2])
    D[0, 0] = 1
    D[0, 1] = -r_left_s
    D[1, 0] = r_right_s
    D[1, 1] = t_right_s*t_left_s - r_right_s*r_left_s
    D = D / t_right_s
    return D

def interface_p(n1, n2, angle1, angle2):
    #n1 is refractive index of leftmost film, while n2 is refracive index of rightmost film

    r_right_p = (n2*np.cos(angle1) - n1*np.cos(angle2))/(n2*np.cos(angle1) + n1*np.cos(angle2))
    r_left_p = (n1*np.cos(angle2) - n2*np.cos(angle1))/(n1*np.cos(angle2) + n2*np.cos(angle1))

    t_right_p = 2*n1*np.cos(angle1)/(n2*np.cos(angle1) + n1*np.cos(angle2))
    t_left_p = 2*n2*np.cos(angle2)/(n1*np.cos(angle2) + n2*np.cos(angle1))

    D = np.ones([2,2])
    D[0, 0] = 1
    D[0, 1] = -r_left_p
    D[1, 0] = r_right_p
    D[1, 1] = t_right_p*t_left_p - r_right_p*r_left_p
    D = D / t_right_p
    return D

def ac_phase(n, d, lam, angle):
    delta = 2*np.pi*n*d*np.cos(angle)/lam
    P = np.array([[0+0j,0+0j], [0+0j,0+0j]])
    P[0,0] = np.exp(1j * delta)
    P[1, 1] = np.exp(-1j * delta)
    return P



from . import animate_data_cube
from . import basic_math
from . import density_scatter_plot
from . import fill_gaps
from . import fpi_sim_matrix_angular
from . import fpi_sim_matrix
from . import fpi_sim
#from . import FPI_TMM
#from . import FPI_class
from . import hsti_export
from . import hsti_import
from . import import_camera_response
from . import import_mirror_sep
#from . import least_squares_methods
from . import lossy_fpi_sim_matrix
from . import mse
#from . import PCA
from . import r_sq
from . import remove_bad_px
from . import remove_vignetting
from . import spectral_debending
from . import voronoi_partitioning


from .animate_data_cube import *
from .basic_math import *
from .density_scatter_plot import *
from .fill_gaps import *
from .fpi_sim_matrix_angular import *
from .fpi_sim_matrix import *
from .fpi_sim import *
from .hsti_export import *
from .hsti_import import *
from .import_camera_response import *
from .import_mirror_sep import *
from .lossy_fpi_sim_matrix import *
from .mse import *
from .r_sq import *
from .remove_bad_px import *
from .remove_vignetting import *
from .spectral_debending import *
from .voronoi_partitioning import *


## Class files
from .PCA import PCA
from .least_squares_methods import ALS,GLS
from .FPI_TMM import Stack
#from .FPI_class import FPI



import numpy as np
import os
import concurrent.futures as cf
import cv2

# Function for saving individual images in .ppm format
def save_img(img, path):
    cv2.imwrite(path, img)
    # Renaming the '.pgm' images to '.ppm' to follow the format coming fron the camera
    new_path = path[0:-2] + 'p' + path[-2+1:]
    os.rename(path, new_path)

def export_data_cube(data_cube, folder_name, normalize_to_cube = True):
    """
    ########## HSTI_export ##########
    This function takes an HSTI numpy array and exports it as individual .ppm
    images to a folder given by folder_name.
    """
    if os.path.isdir(os.getcwd() + '/' + folder_name) == False:
        os.mkdir(os.getcwd() + '/' + folder_name)
        os.mkdir(os.getcwd() + '/' + folder_name + '/images')
        os.mkdir(os.getcwd() + '/' + folder_name + '/images/capture')
    path = os.getcwd() + '/' + folder_name + '/images/capture'
    data_cube = np.rot90(data_cube, 3)
    img_size = [data_cube.shape[0], data_cube.shape[1]]
    NoB = data_cube.shape[2]

    if normalize_to_cube:
        data_cube = (65535.9*(data_cube - np.min(data_cube))/(np.max(data_cube) - np.min(data_cube))).astype(np.uint16)

# setting up for multi processing
    img_lst, path_lst = [], []
    for i in range(NoB):
        img_lst.append(data_cube[:,:,i])
        step = i*10
        path_lst.append(f'{path}/step{step}.pgm')

# Multiprocessing step
    with cf.ProcessPoolExecutor() as executor:
        results = executor.map(save_img, img_lst, path_lst)



import numpy as np
from fnnls import fnnls
from sklearn.isotonic import IsotonicRegression
import numpy.matlib
#########################################################################################################################
#########################################################################################################################
###################### Alternating Least Squares, which returns spectral and contribution profiles ######################
#########################################################################################################################
#########################################################################################################################

## The input matrix is given by X represents spectral measurements. Each row represents a single spectrum.
## X is factorized into a matrix of contributions and a matrix of spectra as well as the error not captured by the model.
## A number of constraints can be set. These include nonnegativity for both spectra and contributions,
## closure can be imposed on the contribution profiles meaning that they will sum to 1.
## Finally, unimodality can be imposed for either the contribution or the spectral profiles such that
## each solution/spectrum only have a single maximum.
## The ALS algorithm requires an initial guess which can be either a spectral profile or contribution profile.
## Two methods are used for this: calculate_ALS_from_spectrum() and calculate_ALS_from_contribution().
## The initial guess determines the number of components, the model will search for in its solution.


class ALS:
# https://doi.org/10.1002/(SICI)1099-128X(199807/08)12:4<223::AID-CEM511>3.0.CO;2-2
# htals.m from hypertools

    def __init__(self):
        self.spectra = 0
        self.contributions = 0
        self.error = []
        self.X = 0
        self.X_hat = 0

## Notes used for previous versions of the code
### hypertools closure:     # self.contributions[j,:] = fnnls(self.spectra.T@self.spectra, self.spectra.T@self.X[j,:])[0] #Taken from hypertools
### hypertools closure:     # self.spectra[j,:] = fnnls(self.contributions.T@self.contributions, self.contributions.T@self.X[:,j])[0] #Taken from hypertools
    def calculate_ALS_from_spectrum(self, matrix, spectra, closure = False, nonnegativity_c = False, nonnegativity_s = False, unimodality_s = False, contrast_weight = 0.00, thresh = 1e-8, max_it = 50):
        self.spectra = np.copy(spectra)
        self.X = np.copy(matrix)
        self.it = 0
        self.contributions = np.zeros([matrix.shape[0], spectra.shape[1]])
        self.temp_specs = []
        for i in range(max_it):
            self.it += 1
            self.temp_specs.append(np.copy(self.spectra))

            self.spectra = (1 - contrast_weight)*self.spectra + contrast_weight*np.matlib.repmat(np.mean(self.spectra, axis = 1 )[:,np.newaxis],1,self.spectra.shape[1])

            if nonnegativity_c:
                for j in range(self.contributions.shape[0]):
                    self.contributions[j,:] = fnnls(self.spectra, self.X[j,:])[0] #self.X[j,:][:,np.newaxis] transforms j'th row of X into column vector


            if closure:
                for j in range(self.contributions.shape[0]):
                    if np.sum(self.contributions[j,:]) > 0:
                        self.contributions[j,:] = self.contributions[j,:]/np.sum(self.contributions[j,:])
                    else:
                        self.contributions[j,:] = np.zeros(self.contributions.shape[1])


            if (closure == False) and (nonnegativity_c == False) and (unimodality_s == False):
                # self.contributions = self.X@self.spectra@np.linalg.pinv(self.spectra.T@self.spectra)
                self.contributions = self.X@np.linalg.pinv(self.spectra.T)


            if nonnegativity_s:
                for j in range(self.spectra.shape[0]):
                    self.spectra[j,:] = fnnls(self.contributions, self.X[:,j])[0]


            if unimodality_s: #https://doi.org/10.1002/(SICI)1099-128X(199807/08)12:4<223::AID-CEM511>3.0.CO;2-2
                ir_incr = IsotonicRegression(increasing=True)
                ir_decr = IsotonicRegression(increasing=False)
                idx = np.arange(self.spectra.shape[0])
                self.c = []
                for j in range(self.spectra.shape[1]): #Loop through each spectrum
                    self.c.append(np.zeros([self.spectra.shape[0],self.spectra.shape[0]]))
                    self.c_fit = []
                    for k in range(self.spectra.shape[0]): #Loop through each possible position of maximum
                        if k == 0:
                            self.c[j][k,k] = self.spectra[k,j]
                            self.c[j][k+1:,k] = ir_decr.fit_transform(idx[k+1:], self.spectra[k+1:,j])
                        elif k == self.spectra.shape[0]-1:
                            self.c[j][:-1,k] = ir_incr.fit_transform(idx[:-1], self.spectra[:-1,j])
                            self.c[j][k,k] = self.spectra[k,j]
                        else:
                            self.c[j][0:k,k] = ir_incr.fit_transform(idx[0:k], self.spectra[0:k,j])
                            self.c[j][k,k] = self.spectra[k,j]
                            self.c[j][k+1:,k] = ir_decr.fit_transform(idx[k+1:], self.spectra[k+1:,j])
                        if (self.c[j][k,k] == np.max(self.c[j][:,k])) and (np.max(self.c[j][:,k]) > 0):
                            self.c_fit.append(self.c[j][:,k])
                    #Find best position of maximum
                    best_fit_err = np.inf
                    best_fit = np.zeros(self.spectra.shape[0])
                    for fit in self.c_fit:
                        if np.linalg.norm(self.spectra[:,j] - fit) < best_fit_err:
                            best_fit = fit
                            best_fit_err = np.linalg.norm(self.spectra[:,j] - fit)
                    self.spectra[:,j] = best_fit


            if (nonnegativity_s == False) and (unimodality_s == False):
                # self.spectra = self.X.T@self.contributions@np.linalg.pinv(self.contributions.T@self.contributions)
                self.spectra = (np.linalg.pinv(self.contributions)@self.X).T

            if np.linalg.norm(self.spectra) > 0:
                self.spectra = self.spectra/np.linalg.norm(self.spectra)
            else:
                self.spectra = np.zeros_like(self.spectra)
            X_old = self.X
            self.X = self.contributions@self.spectra.T
            self.error.append(matrix - self.X)
            if np.linalg.norm(X_old - self.X) < thresh:
                break
            elif i == max_it-1:
                print("Maximum number of iterations reached")

    def calculate_ALS_from_contribution(self, matrix, contributions, closure = False, nonnegativity_c = False, nonnegativity_s = False, unimodality_c = False, thresh = 1e-8, max_it = 50):
        self.contributions = np.copy(contributions)
        self.X = np.copy(matrix)
        self.it = 0
        self.spectra = np.zeros([matrix.shape[1], contributions.shape[1]])
        for i in range(max_it):
            self.it += 1


            if nonnegativity_s:
                for j in range(self.spectra.shape[0]):
                    self.spectra[j,:] = fnnls(self.contributions, self.X[:,j])[0]


            if unimodality_c: #https://doi.org/10.1002/(SICI)1099-128X(199807/08)12:4<223::AID-CEM511>3.0.CO;2-2
                ir_incr = IsotonicRegression(increasing=True)
                ir_decr = IsotonicRegression(increasing=False)
                idx = np.arange(self.contributions.shape[0])
                self.c = []
                for j in range(self.contributions.shape[1]): #Loop through each spectrum
                    self.c.append(np.zeros([self.contributions.shape[0],self.contributions.shape[0]]))
                    self.c_fit = []
                    for k in range(self.contributions.shape[0]): #Loop through each possible position of maximum
                        if k == 0:
                            self.c[j][k,k] = self.contributions[k,j]
                            self.c[j][k+1:,k] = ir_decr.fit_transform(idx[k+1:], self.contributions[k+1:,j])
                        elif k == self.contributions.shape[0]-1:
                            self.c[j][:-1,k] = ir_incr.fit_transform(idx[:-1], self.contributions[:-1,j])
                            self.c[j][k,k] = self.contributions[k,j]
                        else:
                            self.c[j][0:k,k] = ir_incr.fit_transform(idx[0:k], self.contributions[0:k,j])
                            self.c[j][k,k] = self.contributions[k,j]
                            self.c[j][k+1:,k] = ir_decr.fit_transform(idx[k+1:], self.contributions[k+1:,j])
                        if (self.c[j][k,k] == np.max(self.c[j][:,k])) and (np.max(self.c[j][:,k]) > 0):
                            self.c_fit.append(self.c[j][:,k])
                    #Find best position of maximum
                    best_fit_err = np.inf
                    best_fit = np.zeros(self.contributions.shape[0])
                    for fit in self.c_fit:
                        if np.linalg.norm(self.contributions[:,j] - fit) < best_fit_err:
                            best_fit = fit
                            best_fit_err = np.linalg.norm(self.contributions[:,j] - fit)
                    self.contributions[:,j] = best_fit


            if (nonnegativity_s == False) and (unimodality_s == False):
                self.spectra = self.X.T@self.contributions@np.linalg.pinv(self.contributions.T@self.contributions)

            if np.linalg.norm(self.spectra) > 0:
                self.spectra = self.spectra/np.linalg.norm(self.spectra)
            else:
                self.spectra = np.zeros_like(self.spectra)


            if nonnegativity_c:
                for j in range(self.contributions.shape[0]):
                    self.contributions[j,:] = fnnls(self.spectra, self.X[j,:])[0] #self.X[j,:][:,np.newaxis] transforms j'th row of X into column vector


            if closure:
                for j in range(self.contributions.shape[0]):
                    if np.sum(self.contributions[j,:]) > 0:
                        self.contributions[j,:] = self.contributions[j,:]/np.sum(self.contributions[j,:])
                    else:
                        self.contributions[j,:] = np.zeros(self.contributions.shape[1])


            if unimodality_c: #https://doi.org/10.1002/(SICI)1099-128X(199807/08)12:4<223::AID-CEM511>3.0.CO;2-2
                ir_incr = IsotonicRegression(increasing=True)
                ir_decr = IsotonicRegression(increasing=False)
                idx = np.arange(self.contributions.shape[0])
                self.c = []
                for j in range(self.contributions.shape[1]): #Loop through each spectrum
                    self.c.append(np.zeros([self.contributions.shape[0],self.contributions.shape[0]]))
                    self.c_fit = []
                    for k in range(self.contributions.shape[0]): #Loop through each possible position of maximum
                        if k == 0:
                            self.c[j][k,k] = self.contributions[k,j]
                            self.c[j][k+1:,k] = ir_decr.fit_transform(idx[k+1:], self.contributions[k+1:,j])
                        elif k == self.contributions.shape[0]-1:
                            self.c[j][:-1,k] = ir_incr.fit_transform(idx[:-1], self.contributions[:-1,j])
                            self.c[j][k,k] = self.contributions[k,j]
                        else:
                            self.c[j][0:k,k] = ir_incr.fit_transform(idx[0:k], self.contributions[0:k,j])
                            self.c[j][k,k] = self.contributions[k,j]
                            self.c[j][k+1:,k] = ir_decr.fit_transform(idx[k+1:], self.contributions[k+1:,j])
                        if (self.c[j][k,k] == np.max(self.c[j][:,k])) and (np.max(self.c[j][:,k]) > 0):
                            self.c_fit.append(self.c[j][:,k])
                    #Find best position of maximum
                    best_fit_err = np.inf
                    best_fit = np.zeros(self.contributions.shape[0])
                    for fit in self.c_fit:
                        if np.linalg.norm(self.contributions[:,j] - fit) < best_fit_err:
                            best_fit = fit
                            best_fit_err = np.linalg.norm(self.contributions[:,j] - fit)
                    self.contributions[:,j] = best_fit


            if (closure == False) and (nonnegativity_c == False) and (unimodality_c == False):
                self.contributions = self.X@self.spectra@np.linalg.pinv(self.spectra.T@self.spectra)

            X_old = self.X
            self.X = self.contributions@self.spectra.T
            self.error.append(matrix - self.contributions@self.spectra.T)
            if np.linalg.norm(X_old - self.X) < thresh:
                break
            elif i == max_it-1:
                print("Maximum number of iterations reached")


########################################################################################################################
########################################################################################################################
############################################## Generalized Least Squares  ##############################################
########################################################################################################################
########################################################################################################################


class GLS:
    def __init__(self):
        self.spectra = 0
        self.contributions = 0
        self.error = 0
        self.X = 0
        self.X_hat = 0
        self.lam = 0

## Calculate the contributions of target and clutter. The data matrix is structured
## to have each sample represented by a row with each column representing a single wavelength.
## A target spectrum is supplied, and clutter is then defined as the mean of all the data.
## The data covariance matrix is used to deweigh the most prominent directions within the data
## such that the target stands out more easily. For this deweighting, the data is regularized.
## This is done by adding a constant value to the eigen values of the clutter covariance matrix.
## This value is determined by the regularization_parameter

## The contributions contain two vectors. The first describe the contribution of the targeted
## while the second describe the contribution of the clutter.
    def calculate_contributions(self, matrix, target, clutter = None, regularization_parameter = 1e-5):
        self.X = np.copy(matrix)
        self.target = np.copy(target)
        if clutter == None:
            clutter = np.mean(self.X, axis = 0)[:,np.newaxis]
        self.spectra = np.c_[self.target, clutter]
        self.lam = regularization_parameter
        self.weights = np.cov(self.X.T)

        U, s, Vh = np.linalg.svd(self.weights, full_matrices = False)
        S = np.diag(s)
        D = np.eye(S.shape[0])*self.lam

        self.weights_inv = Vh.T@np.linalg.inv(S+D)@U.T

        self.contributions = np.linalg.inv(self.spectra.T@self.weights_inv@self.spectra)@self.spectra.T@self.weights_inv@self.X.T
        self.X_hat = self.contributions.T@self.spectra.T
        self.error = self.X - self.X_hat



import numpy as np
from IPython.display import clear_output
import concurrent.futures as cf

def fpi_trans_single_mirror_sep(wavelength, temp, layer_t):
    transmittance_vec = np.zeros_like(wavelength)
    for idx, wvl in enumerate(wavelength):
        stack_refractive_index = np.array([refrac_Ge(wvl, temp), refrac_ThF4(wvl), refrac_Ge(wvl, temp), 1, refrac_Ge(wvl, temp), refrac_ThF4(wvl), refrac_Ge(wvl, temp)])
        T = transfer_FPI(stack_refractive_index, layer_t, wvl)
        transmittance_vec[idx] = abs(1/T[0,0])**2
    return transmittance_vec

def FPI_trans_matrix(mirror_sep, lam, temperature):
    lam0 = 10.5e-6 #Central wavelength used for layer thicknesses

    transmittance = np.zeros([len(mirror_sep), len(lam)])
    layer_thickness = np.array([0.5*lam0/refrac_Ge(lam0, temperature), 0.25*lam0/refrac_ThF4(lam0),  0.25*lam0/refrac_Ge(lam0, temperature), 0, 0.25*lam0/refrac_Ge(lam0, temperature), 0.25*lam0/refrac_ThF4(lam0), 0.5*lam0/refrac_Ge(lam0, temperature)])

    #Formulate input into lists for later multiprocessing
    lam_lst, temperature_lst, layer_thickness_lst, = [], [], []
    for m_sep in mirror_sep:
        lam_lst.append(lam)
        temperature_lst.append(temperature)
        temp_layer_thickness = np.copy(layer_thickness)
        temp_layer_thickness[3] = m_sep
        layer_thickness_lst.append(temp_layer_thickness)


    with cf.ProcessPoolExecutor() as executor:
        results = executor.map(fpi_trans_single_mirror_sep, lam_lst, temperature_lst, layer_thickness_lst)

    for i, result in enumerate(results):
        transmittance[i,:] = result
    return transmittance

def refrac_ThF4(lam):
    # Handbook of Optical Constants of Solids II, 1998, page 1051
    lam = lam * 1e6
    n = 1.118 + 2.46 / lam - 2.98 / (lam**2)
    return n

def refrac_ZnSe(lam):
    # Handbook of Optical Constants of Solids II, page 737
    lam = lam * 1e6
    n = np.sqrt(4 + 1.9*lam**2/(lam**2 - 0.113))
    return n

def refrac_Ge(lam, temperature):
    # Handbook of Optical Constants of Solids I, page 465
    lam = lam * 1e6
    A = -6.04e-3 * temperature + 11.05128
    B = 9.295e-3 * temperature + 4.00536
    C = -5.392e-4 * temperature + 0.599034
    D = 4.151e-4 * temperature + 0.09145
    E = 1.51408 * temperature + 3426.5
    n = np.sqrt(A + (B * lam**2) / (lam**2 - C) + (D * lam**2) / (lam**2 - E))
    return n

def transfer_FPI(n, d, lam):
    T = interface(refrac_ZnSe(lam), n[0]) @ ac_phase(n[0], d[0], lam)
    for i in np.arange(1, len(n), 1):
        T = T @ interface(n[i-1], n[i]) @ ac_phase(n[i], d[i], lam)
    T = T @ interface(n[-1], refrac_ZnSe(lam))
    return T

def interface(n1, n2):
    #n1 is refractive index of leftmost film, while n2 is refracive index of rightmost film
    D = np.ones([2,2])
    D[0, 0] = 1 + n2 / n1
    D[0, 1] = 1 - n2 / n1
    D[1, 0] = 1 - n2 / n1
    D[1, 1] = 1 + n2 / n1
    D = 0.5 * D
    return D

def ac_phase(n, d, lam):
    P = np.array([[0+0j,0+0j], [0+0j,0+0j]])
    P[0,0] = np.exp(1j * n * 2 * np.pi * d / lam)
    P[1, 1] = np.exp(-1j * n * 2 * np.pi * d / lam)
    return P



import numpy as np
from scipy.ndimage import median_filter
from IPython.display import clear_output
import concurrent.futures as cf
import os
import cv2 as cv

#Multiple of these functions takes an optional argument defining in which axis
#the function operates. If 's' is passed as an argument, the operation is performed
#along the spectral axis (across bands/variables). If the argument is 'b', the
#function operates on each band/variable individually

#Subtrats the mean from the data, either the mean of each spectrum (axis = 's')
#or the mean of each band (axis = 'b')
def mean_center(data_cube, axis = 's'):
    if axis == 's':
        return data_cube - np.mean(data_cube, axis = 2)[:,:,np.newaxis]
    elif axis == 'b':
        return data_cube - np.mean(data_cube, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Subtracts mean and scales with STD - the old standardize function. Setting
#axis = 's' is the same as doing SNV (standard normal variate)
def autoscale(data_cube, axis = 's'):
    if axis == 's':
        std_cube = np.std(data_cube, axis = 2)[:,:,np.newaxis]
        std_cube[std_cube < 1e-6] = np.nan
        return (data_cube - np.mean(data_cube, axis = 2)[:,:,np.newaxis])/std_cube
    elif axis == 'b':
        return (data_cube - np.mean(data_cube, axis = (0,1)))/np.std(data_cube, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Uses norm of given order for normalization. If axis = 's', then each spectrum is
#divided by its norm. If axis = 'b', then every band is divided by the norm of
#the entire band.
def norm_normalization(data_cube, order, axis = 's'):
    if axis == 's':
        return data_cube/np.linalg.norm(data_cube, ord = order, axis = 2)[:,:,np.newaxis]
    elif axis == 'b':
        return data_cube/np.linalg.norm(data_cube, ord = order, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Apply multiplicative scatter correction to entire datacube. If no reference
#spectrum is supplied, the mean spectrum of the cube is used instead.
def msc(data_cube, ref_spec = None):
    if ref_spec == None:
        ref_spec = np.mean(data_cube, axis = (0,1))
        ref_spec_mean = np.mean(ref_spec)
        ref_spec = ref_spec - ref_spec_mean
    else:
        ref_spec_mean = np.mean(ref_spec)
        ref_spec - ref_spec_mean
    ref_spec = ref_spec[:,np.newaxis]
    X = mean_center(data_cube, 's')
    X = X.reshape([X.shape[0]*X.shape[1], X.shape[2]])
    X = np.rot90(X, 3)

    b = np.linalg.inv(ref_spec.T@ref_spec)@ref_spec.T@X
    b[b<1e-8] = np.nan
    X_new = X/b + ref_spec_mean
    X_new = np.rot90(X_new)
    X_new = X_new.reshape([data_cube.shape[0], data_cube.shape[1], data_cube.shape[2]])
    return X_new

#Setting axis = 's' is the same as normalizing each spectrum (pixel) to span from 0 to 1
#axis = 'b' normalizes each band individually
def normalize(data_cube, axis = 's'):
    if axis == 's':
        cube = data_cube - np.min(data_cube, axis = 2)[:,:,np.newaxis]
        max_cube = np.max(cube, axis = 2)[:,:,np.newaxis]
        return cube/max_cube
    elif axis == 'b':
        cube = data_cube - np.min(data_cube, axis = (0,1))
        return cube/np.max(data_cube, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Normalizes the cube to span from 0 to 1
def normalize_cube(data_cube):
    cube = data_cube - np.min(data_cube)
    return cube/np.max(cube)

#Subtracts selected band from all layers in the cube
def subtract_band(data_cube, band):
    return data_cube - data_cube[:,:,band][:,:,np.newaxis]

#Flattens data cube into 2D array
def flatten(data_cube):
    return np.reshape(data_cube, [data_cube.shape[0]*data_cube.shape[1], data_cube.shape[2]])

#Reshapes flattened data cube back into 3 dimensions
def inflate(array_2D, n_rows, n_cols):
    return np.reshape(array_2D, [n_rows, n_cols, matrix.shape[1]])

#Median filters each band in data cube
def median_filter_cube(data_cube, kernel_size):
    band_lst = []
    for i in range(data_cube.shape[2]):
        band_lst.append(data_cube[:,:,i])
    kernel_size_lst = np.ones(len(band_lst), dtype = int)*kernel_size
    with cf.ThreadPoolExecutor() as executor:
        results = executor.map(median_filter, band_lst, kernel_size_lst)
    filtered_cube = np.zeros_like(data_cube)
    for i, result in enumerate(results):
        filtered_cube[:,:,i] = result
    return filtered_cube

#Median filters a 2D array and only applies it to the locations marked as True
#in the px_idx array.
def targeted_median_filter(input_array, px_idx, kernel_size):
    if input_array.shape != px_idx.shape:
        print("Arrays must be the same shape. px_idx must be a true/false numpy array")
        return
    if type(input_array[0,0]) is not np.float64:
         input_array = input_array.astype(float)
    filtered_array = median_filter(input_array, size = kernel_size)
    input_array[px_idx] = filtered_array[px_idx]
    return input_array

#Imports cube based on directory argument
def load_cube(directory):
    NoB = sum([ '.ppm' in s for s in os.listdir(f'{directory}images/capture')])
    steps = np.arange(0,NoB*10,10) #vector with all the steps at which an image is taken
    cube = []
    for step in steps: #load bands into data structure
        cube.append(np.rot90(cv.imread(f'{directory}images/capture/step{step}.ppm',cv.IMREAD_ANYDEPTH)))
    cube = np.array(cube, dtype = 'float64')
    cube = np.moveaxis(cube, 0, 2) #Rearranges the array to required shape
    return cube

#Calculates hottelings T^2 statistic for matrix X, where the variables are
#represented by each column and the samples by the rows
#https://learnche.org/pid/latent-variable-modelling/principal-component-analysis/hotellings-t2-statistic
def hottelings(X):
    return np.sum((X/np.std(X, axis=0))**2, axis=1)

#Calculates the upper and lower 95% confidence limits of the mean of input vector, x
def conf95lim(x):
    conf = []
    mean = np.mean(x)
    std = np.std(x)
    conf.append(mean-2*std)
    conf.append(mean+2*std)
    return conf

#Select three layers (either as a list or numpy vector) and use these three as the
#channels of an rgb image. The first layer is the red channel, the second layer
#the green channel and the third layer is the blue channel.
def array2rgb(data_cube, three_layers):
    three_layer_cube = np.zeros([data_cube.shape[0], data_cube.shape[1], 3])
    three_layer_cube[:,:,0] = data_cube[:,:,three_layers[0]] - np.min(data_cube[:,:,three_layers[0]])
    three_layer_cube[:,:,0] = three_layer_cube[:,:,0]/np.max(three_layer_cube[:,:,0])
    three_layer_cube[:,:,1] = data_cube[:,:,three_layers[1]] - np.min(data_cube[:,:,three_layers[1]])
    three_layer_cube[:,:,1] = three_layer_cube[:,:,1]/np.max(three_layer_cube[:,:,1])
    three_layer_cube[:,:,2] = data_cube[:,:,three_layers[2]] - np.min(data_cube[:,:,three_layers[2]])
    three_layer_cube[:,:,2] = three_layer_cube[:,:,2]/np.max(three_layer_cube[:,:,2])
    return three_layer_cube

#This function calculates and applies a NUC to the entire datacube. The NUC is
#dependent on the sensor temperature and the GSK settings of the camera. The
#NUC is calculated from camera specific calibration files from the accompanying NUC directory
def apply_NUC_cube(cube, sensor_temp, GSK, NUC_directory = '/home/mads/git/thermal-camera-src/src/python3-scripts/hsti python module/hsti-analysis/HSTI/data_files/NUC/10_10_200_191/'):
    offset_coefs = np.load(f'{NUC_directory}offset_coefs_1st_order.npy', allow_pickle = True)
    slope_coefs = np.load(f'{NUC_directory}slope_coefs_1st_order.npy', allow_pickle = True)

    M = np.array([[GSK, sensor_temp, 1]]).T

    offsets = offset_coefs@M
    slopes = slope_coefs@M

    mean_offset = np.nanmean(offsets)
    mean_slopes = np.nanmean(slopes)

    offset_correction = - mean_slopes*offsets/slopes + mean_offset
    slope_correction = (mean_slopes - slopes)/slopes

    offset_matrix = offset_correction.reshape([cube.shape[0], cube.shape[1]], order = 'C')
    slope_matrix = slope_correction.reshape([cube.shape[0], cube.shape[1]], order = 'C')

    temp_cube = np.copy(cube)
    for i in range(temp_cube.shape[2]):
        temp_cube[:,:,i] = temp_cube[:,:,i] + temp_cube[:,:,i]*slope_matrix + offset_matrix
    return temp_cube

#This function calculates and applies a NUC to a single image. The NUC is dependent
#on the sensor temperature and the GSK settings of the camera. The NUC is
#calculated from camera specific calibration files from the accompanying NUC directory.
def apply_NUC_image(image, sensor_temp, GSK, NUC_directory = '/home/mads/git/thermal-camera-src/src/python3-scripts/hsti python module/hsti-analysis/HSTI/data_files/NUC/10_10_200_191/'):
    offset_coefs = np.load(f'{NUC_directory}offset_coefs_1st_order.npy', allow_pickle = True)
    slope_coefs = np.load(f'{NUC_directory}slope_coefs_1st_order.npy', allow_pickle = True)

    M = np.array([[GSK, sensor_temp, 1]]).T

    offsets = offset_coefs@M
    slopes = slope_coefs@M

    mean_offset = np.nanmean(offsets)
    mean_slopes = np.nanmean(slopes)

    offset_correction = - mean_slopes*offsets/slopes + mean_offset
    slope_correction = (mean_slopes - slopes)/slopes

    offset_matrix = offset_correction.reshape([image.shape[0], image.shape[1]], order = 'C')
    slope_matrix = slope_correction.reshape([image.shape[0], image.shape[1]], order = 'C')

    return image  + image*slope_matrix + offset_matrix



import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from fast_histogram import histogram2d

def density_scatter_plot(X, Y, n_bins = 100, colormap = 'plasma', ax = None):
    if ax is None :
        fig , ax = plt.subplots()

    cmap = plt.get_cmap(colormap)
    bounds = [[np.nanmin(X), np.nanmax(X)], [np.nanmin(Y), np.nanmax(Y)]]
    h = histogram2d(X, Y, range=bounds, bins=n_bins)
    h = np.rot90(h)
    density = ax.imshow(h, extent=[bounds[0][0],bounds[0][1],bounds[1][0],bounds[1][1]], aspect='auto',\
               norm=colors.LogNorm(vmin=1, vmax=h.max()), cmap=cmap, interpolation = 'none')
    fig.colorbar(density, label='Number of points per pixel')
    return fig, ax



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="HSTI",
    version="0.0.39",
    author="Mads Nibe et. al.",
    author_email="mani@newtec.dk",
    description="The NEWTEC HSTI package contains fundamental functions for the data analysis of hyperspectral thermal images (HSTI).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_data={'': ['data_files/*.txt']},
    package_dir={"": "hsti-analysis"},
    packages=setuptools.find_packages(where="hsti-analysis"),
    python_requires=">=3.6",
    install_requires=[
        "fast-histogram>=0.10",
        "fnnls>=1.0.0",
        "ipython>=7.25.0",
        "jupyter",
        "matplotlib>=3.4.2",
        "natsort>=7.1.1",
        "notebook>=6.4.0",
        "numpy>=1.19.5",
        "opencv-python>=4.5.2.54",
        "Pillow>=8.4.0",
        "scikit-learn>=0.24.2",
        "scipy>=1.7.0",
        "Screeninfo>=0.8",
        "RangeSlider>=2021.7.4",
        "tk>=0.1.0"
    ]
)



import numpy as np
import os

def import_mirror_sep_from_path(path, number_of_bands):
    t = np.arange(0,number_of_bands,1)
    p = np.poly1d(np.load(path))
    mirror_sep = p(t)
    if np.max(mirror_sep) > 100e-6:
        mirror_sep = mirror_sep*1e-6
    return mirror_sep



import numpy as np

def FPI_trans(mirror_sep,lam, temp):
    lam0 = 10.5e-6
    layer_thickness = np.array([0.5*lam0/refrac_Ge(lam0, temp), 0.25*lam0/refrac_ThF4(lam0),  0.25*lam0/refrac_Ge(lam0, temp), mirror_sep, 0.25*lam0/refrac_Ge(lam0, temp), 0.25*lam0/refrac_ThF4(lam0), 0.5*lam0/refrac_Ge(lam0, temp)])
    #layer_thickness = np.array([0.5*lam0/refrac_Ge(lam0, temp), 1682e-9,  0.25*lam0/refrac_Ge(lam0, temp), mirror_sep, 0.25*lam0/refrac_Ge(lam0, temp), 1682e-9, 0.5*lam0/refrac_Ge(lam0, temp)])
    stack_refractive_index = np.array([refrac_Ge(lam, temp), refrac_ThF4(lam), refrac_Ge(lam, temp), 1, refrac_Ge(lam, temp), refrac_ThF4(lam), refrac_Ge(lam, temp)])
    T = transfer_FPI(stack_refractive_index, layer_thickness, lam)
    reflectance = np.abs(T[1,0]/T[0,0])**2
    transmittance = abs(1/T[0,0])**2

    loss = 1 - transmittance - reflectance
    return transmittance, reflectance, loss

def refrac_ThF4(lam):
    # Handbook of Optical Constants of Solids II, 1998, page 1051
    lam = lam * 1e6
    n = 1.118 + 2.46 / lam - 2.98 / (lam**2)
    return n

def refrac_ZnSe(lam):
    # Handbook of Optical Constants of Solids II, page 737
    lam = lam * 1e6
    n = np.sqrt(4 + 1.9*lam**2/(lam**2 - 0.113))
    return n

def refrac_Ge(lam, temp):
    # Handbook of Optical Constants of Solids I, page 465
    lam = lam * 1e6
    A = -6.04e-3 * temp + 11.05128
    B = 9.295e-3 * temp + 4.00536
    C = -5.392e-4 * temp + 0.599034
    D = 4.151e-4 * temp + 0.09145
    E = 1.51408 * temp + 3426.5
    n = np.sqrt(A + (B * lam**2) / (lam**2 - C) + (D * lam**2) / (lam**2 - E))
    return n

def transfer_FPI(n, d, lam):
    T = interface(refrac_ZnSe(lam), n[0]) @ ac_phase(n[0], d[0], lam)
    for i in np.arange(1, len(n), 1):
        T = T @ interface(n[i-1], n[i]) @ ac_phase(n[i], d[i], lam)
    T = T @ interface(n[-1], refrac_ZnSe(lam))
    return T

def interface(n1, n2):
    #n1 is refractive index of leftmost film, while n2 is refracive index of rightmost film
    D = np.ones([2,2])
    D[0, 0] = 1 + n2 / n1
    D[0, 1] = 1 - n2 / n1
    D[1, 0] = 1 - n2 / n1
    D[1, 1] = 1 + n2 / n1
    D = 0.5 * D
    return D

def ac_phase(n, d, lam):
    P = np.array([[0+0j,0+0j], [0+0j,0+0j]])
    P[0,0] = np.exp(1j * n * 2 * np.pi * d / lam)
    P[1, 1] = np.exp(-1j * n * 2 * np.pi * d / lam)
    return P
